﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using HeuristicLab.Algorithms.GeneticAlgorithm;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.CombinedIntegerVectorEncoding;
using HeuristicLab.Encodings.IntegerVectorEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Optimization.Operators;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Random;
using HeuristicLab.Selection;

namespace HeuristicLab.Algorithms.LearningClassifierSystems {
  /// <summary>
  /// An operator which represents the main loop of a learning classifier system.
  /// </summary>
  [Item("LearningClassifierSystemMainLoop", "An operator which represents the main loop of a learning classifier system.")]
  [StorableClass]
  public class LearningClassifierSystemMainLoop : AlgorithmOperator {

    #region Parameter Properties
    public IConstrainedValueParameter<ISelector> SelectorParameter {
      get { return (IConstrainedValueParameter<ISelector>)Parameters["Selector"]; }
    }
    public IConstrainedValueParameter<ICrossover> CrossoverParameter {
      get { return (IConstrainedValueParameter<ICrossover>)Parameters["Crossover"]; }
    }
    public IConstrainedValueParameter<IManipulator> MutatorParameter {
      get { return (IConstrainedValueParameter<IManipulator>)Parameters["Mutator"]; }
    }
    #endregion

    #region Properties
    public ISelector Selector {
      get { return SelectorParameter.Value; }
      set { SelectorParameter.Value = value; }
    }
    public ICrossover Crossover {
      get { return CrossoverParameter.Value; }
      set { CrossoverParameter.Value = value; }
    }
    public IManipulator Mutator {
      get { return MutatorParameter.Value; }
      set { MutatorParameter.Value = value; }
    }
    #endregion

    [StorableConstructor]
    private LearningClassifierSystemMainLoop(bool deserializing) : base(deserializing) { }
    protected LearningClassifierSystemMainLoop(LearningClassifierSystemMainLoop original, Cloner cloner)
      : base(original, cloner) {
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new LearningClassifierSystemMainLoop(this, cloner);
    }
    public LearningClassifierSystemMainLoop()
      : base() {
      Initialize();
    }

    private void Initialize() {
      #region Create parameters
      Parameters.Add(new ConstrainedValueParameter<ISelector>("Selector", "The operator used to select solutions for reproduction.", new ItemSet<ISelector>() { new ProportionalSelector() }, new ProportionalSelector()));
      Parameters.Add(new ConstrainedValueParameter<ICrossover>("Crossover", "The operator used to cross solutions.", new ItemSet<ICrossover>() { new SinglePointCrossover() }, new SinglePointCrossover()));
      Parameters.Add(new OptionalConstrainedValueParameter<IManipulator>("Mutator", "The operator used to mutate solutions.", new ItemSet<IManipulator>() { new UniformOnePositionManipulator() }, new UniformOnePositionManipulator()));
      //for test purposes
      int[] numbers = new int[3] { 1, 1, 1 };
      int[,] elements = new int[,] { { 0, 3 }, { 0, 3 }, { 0, 2 } };
      Parameters.Add(new ValueLookupParameter<CombinedIntegerVector>("TestTarget", "Target for test", new CombinedIntegerVector(numbers, 1, elements)));

      #endregion

      #region Create operators
      VariableCreator variableCreator = new VariableCreator();
      MatchSelector conditionMatchSelector = new MatchSelector();
      MatchSelector actionMatchSelector = new MatchSelector();
      SubScopesProcessor coveringSubScopesProcessor = new SubScopesProcessor();
      SubScopesRemover subScopesRemover = new SubScopesRemover();
      Comparator comparator = new Comparator();
      ConditionalBranch conditionalBranch = new ConditionalBranch();
      MergingReducer mergingReducer = new MergingReducer();
      SolutionsCreator coveringMechanism = new SolutionsCreator();
      SubScopesProcessor actionSelectionSubScopesProcessor = new SubScopesProcessor();
      MaxValueActionSelector maxValueActionSelection = new MaxValueActionSelector();
      GeneticAlgorithmMainLoop geneticAlgorithmMainLoop = new GeneticAlgorithmMainLoop();

      UniformSubScopesProcessor uniformSubScopeProcessor = new UniformSubScopesProcessor();
      UniformRandomizer uniformRandomizer = new UniformRandomizer();

      variableCreator.CollectedValues.Add(new ValueParameter<IntValue>("ZeroIntValue", new IntValue(0)));
      variableCreator.CollectedValues.Add(new ValueParameter<IntValue>("OneIntValue", new IntValue(1)));

      conditionMatchSelector.CopySelected = new BoolValue(false);
      conditionMatchSelector.MatchCondition = new BoolValue(true);
      //this has to be set differently
      conditionMatchSelector.MatchParameter.ActualName = "CombinedIntegerVector";
      conditionMatchSelector.TargetMatchParameter.ActualName = "TestTarget";

      comparator.Comparison = new Comparison(ComparisonType.LessOrEqual);
      comparator.LeftSideParameter.ActualName = conditionMatchSelector.NumberOfSelectedSubScopesParameter.ActualName;
      comparator.ResultParameter.ActualName = "Covering";
      //more than zero solutions have to be selected
      comparator.RightSideParameter.ActualName = "ZeroIntValue";

      conditionalBranch.ConditionParameter.ActualName = "Covering";

      subScopesRemover.SubScopeIndexParameter.ActualName = "OneIntValue";
      subScopesRemover.RemoveAllSubScopes = false;

      coveringMechanism.NumberOfSolutionsParameter.ActualName = "OneIntValue";
      coveringMechanism.SolutionCreatorParameter.ActualName = "SolutionCreator";

      maxValueActionSelection.ValueParameter.ActualName = "Fitness";
      maxValueActionSelection.MatchParameter.ActualName = "CombinedIntegerVector";

      actionMatchSelector.CopySelected = new BoolValue(false);
      actionMatchSelector.MatchCondition = new BoolValue(false);
      //this has to be set differently
      actionMatchSelector.MatchParameter.ActualName = "CombinedIntegerVector";
      actionMatchSelector.TargetMatchParameter.ActualName = maxValueActionSelection.SelectedActionParameter.ActualName;

      SelectorParameter.Value.CopySelected = new BoolValue(true);
      SelectorParameter.Value.NumberOfSelectedSubScopesParameter.Value = new IntValue(200);

      geneticAlgorithmMainLoop.SelectorParameter.ActualName = SelectorParameter.Name;
      geneticAlgorithmMainLoop.CrossoverParameter.ActualName = CrossoverParameter.Name;
      geneticAlgorithmMainLoop.MutatorParameter.ActualName = MutatorParameter.Name;
      geneticAlgorithmMainLoop.RandomParameter.ActualName = "Random";
      geneticAlgorithmMainLoop.ElitesParameter.ActualName = "OneIntValue";
      geneticAlgorithmMainLoop.MaximumGenerationsParameter.ActualName = "ZeroIntValue";
      geneticAlgorithmMainLoop.QualityParameter.ActualName = "Fitness";
      //needed?
      geneticAlgorithmMainLoop.MutationProbabilityParameter.Value = new PercentValue(10);
      geneticAlgorithmMainLoop.MaximizationParameter.Value = new BoolValue(true);

      uniformSubScopeProcessor.Parallel = new BoolValue(true);

      uniformRandomizer.RandomParameter.ActualName = "Random";
      uniformRandomizer.MinParameter.Value = new DoubleValue(0);
      uniformRandomizer.MaxParameter.Value = new DoubleValue(100);
      uniformRandomizer.ValueParameter.ActualName = "Fitness";

      //ParameterizeSelectors();
      #endregion

      #region Create operator graph
      OperatorGraph.InitialOperator = variableCreator;
      variableCreator.Successor = conditionMatchSelector;
      //variableCreator.Successor = geneticAlgorithmMainLoop;
      //geneticAlgorithmMainLoop.Successor = conditionMatchSelector;
      conditionMatchSelector.Successor = comparator;
      comparator.Successor = conditionalBranch;
      conditionalBranch.TrueBranch = coveringSubScopesProcessor;
      conditionalBranch.FalseBranch = actionSelectionSubScopesProcessor;
      conditionalBranch.Successor = subScopesRemover;
      coveringMechanism.Successor = uniformSubScopeProcessor;
      uniformSubScopeProcessor.Operator = uniformRandomizer;
      coveringSubScopesProcessor.Operators.Add(new EmptyOperator());
      coveringSubScopesProcessor.Operators.Add(coveringMechanism);
      coveringSubScopesProcessor.Successor = actionSelectionSubScopesProcessor;
      actionSelectionSubScopesProcessor.Operators.Add(new EmptyOperator());
      actionSelectionSubScopesProcessor.Operators.Add(maxValueActionSelection);
      maxValueActionSelection.Successor = actionMatchSelector;
      subScopesRemover.Successor = mergingReducer;
      mergingReducer.Successor = conditionMatchSelector;
      //mergingReducer.Successor = geneticAlgorithmMainLoop;
      #endregion
    }

    private void ParameterizeStochasticOperator(IOperator op) {
      IStochasticOperator stochasticOp = op as IStochasticOperator;
      if (stochasticOp != null) {
        stochasticOp.RandomParameter.ActualName = "Random";
        stochasticOp.RandomParameter.Hidden = true;
      }
    }
    //private void ParameterizeSelectors() {
    //  foreach (ISelector selector in SelectorParameter.ValidValues) {
    //    selector.CopySelected = new BoolValue(true);
    //    //set value by parameter!
    //    selector.NumberOfSelectedSubScopesParameter.Value = new IntValue(5);
    //    selector.NumberOfSelectedSubScopesParameter.Hidden = true;
    //    ParameterizeStochasticOperator(selector);
    //  }
    //  if (Problem != null) {
    //    foreach (ISingleObjectiveSelector selector in SelectorParameter.ValidValues.OfType<ISingleObjectiveSelector>()) {
    //      selector.MaximizationParameter.Value = new BoolValue(true);
    //      selector.MaximizationParameter.Hidden = true;
    //      selector.QualityParameter.ActualName = Problem.Evaluator.QualityParameter.ActualName;
    //      selector.QualityParameter.Hidden = true;
    //    }
    //  }
    //}
  }
}
